using System;
using System.Collections;
using System.ComponentModel;
using System.Diagnostics;
using System.Drawing;
using System.Drawing.Design;
using System.Windows.Forms;

using Infragistics.UltraChart.Core.ColorModel; // for IColorModel
using Infragistics.UltraChart.Core;
using Infragistics.UltraChart.Core.Layers;
using Infragistics.UltraChart.Core.Primitives;
using Infragistics.UltraChart.Data;
using Infragistics.UltraChart.Resources;   // for IChartComponent
using Infragistics.UltraChart.Resources.Appearance;
using Infragistics.UltraChart.Resources.Editor;
using Infragistics.UltraChart.Shared.Styles;

namespace CSFinancialTrendLayer
{
	/// <summary>
	/// This layer renders the moving average overlays.
	/// 
	/// I extend Layer, which provides basic ILayer state management.
	/// The Layer Developer's primary responsibility is to implement
	/// a FillSceneGraph() method, and wire the Layer into UltraChart's
	/// Layers collection when appropriate.
	/// </summary>
	public class TrendLayer : Layer
	{
		private uint daysMovAvg;
		private DataPoint[] movAvg;
		private Color mavgColor;

		/// <summary>
		/// This method creates a TrendLayer based on moving averages of
		/// 'days' data samples.
		/// 
		/// I call Layer's constructor with null arguments.  As long
		/// as I invoke UltraWinChart (or UltraWebChart's) InvalidateLayers()
		/// method when activating a Layer in the Layers Collection, the
		/// ChartComponent and ChartCore properties will be properly
		/// initialized before any painting is performed.
		/// </summary>
		public TrendLayer( int days) : base( null, null)
		{
			this.movAvg = null;

			// It's not possible to validate a sufficient numbers of days
			// has been specified right now; IChartData may not be specified
			// until FSG() is called.
			this.daysMovAvg = (uint) days;

			// Set a contrasting color for expected SMA periods.
			switch ( days )
			{
				case 5:
					mavgColor = Color.DodgerBlue;
					break;
				case 10:
					mavgColor = Color.Fuchsia;
					break;
				case 20:
					mavgColor = Color.Orange;
					break;
				default:
					mavgColor = Color.FromArgb( days & 0x38 << 2, days & 0x0f << 4, days & 0x38 << 2);
					break;
			}
		}

		/// <summary>
		/// The number of 'days' the Moving Average of this Layer was
		/// calculated on.  Days may mean hours or minutes, depending
		/// on the frequency of data samples.  Days must always be
		/// greater than the number of data samples (the layer must
		/// not render an unstable moving average).
		/// </summary>
		public uint Days
		{
			get
			{
				return this.daysMovAvg;
			}

			set
			{
				if ( this.isDaysValid( value))
				{
					this.daysMovAvg = value;
					this.movAvg = new DataPoint[ this.GetData().GetRowCount() - value + 1];
				}
				else
				{
					throw new ArgumentOutOfRangeException( "value", value, String.Format("Days must be set between {0} and {1} for valid {0}-day Moving Average calculation.", this.daysMovAvg, this.GetData().GetRowCount()));
				}
			}
		}

		private bool isDaysValid( uint prospectValue)
		{
			// Validate the property value assigned to Days is not
			// longer than the historical data available to chart.
			int runningLength = (int)prospectValue;

			try
			{
				runningLength = this.GetData().GetRowCount() - runningLength;
			}
			catch ( NullReferenceException )
			{
				// Called prior to IChartData being initialized,
				// therefore validation is not possible.
				;
			}

			return ( runningLength > 0 );
		}

		private void InitAdvanceAxis( ref IAdvanceAxis x, ref IAdvanceAxis y)
		{
			x = this.Grid["X"] as IAdvanceAxis;
			y = this.Grid["Y"] as IAdvanceAxis;
			
			x.MapMinimum = this.OuterBound.Left;
			x.MapMaximum = this.OuterBound.Right;
			y.MapMinimum = this.OuterBound.Bottom;
			y.MapMaximum = this.OuterBound.Top;

			x.Minimum = 0;
			x.Maximum = this.GetData().GetRowCount() - 1;
			y.Minimum = this.GetData().GetDataMinColumn( 4);
			y.Maximum = this.GetData().GetDataMaxColumn( 4);

			return;
		}

		/// <summary>
		/// The default implementation in Layer does nothing, so developers must
		/// provide the visual rendition of their custom layer in this method's body.
		/// </summary>
		/// <param name="scene">The scene to render to will be provided.</param>
		public override void FillSceneGraph( SceneGraph scene)
		{
			// Verify there is sufficient data to graph (this may
			// be the first time I have data.)
			try
			{
				Days = Days;
			}
			catch ( ArgumentOutOfRangeException)
			{
				return;
			}

			// Calculate the Simple (or Arithmetic) Moving Average.
			calcSMA( this.movAvg);

			// Create a Polyline object and fill in the MA DataPoint[] 
			// with the default Appearance set up for a LineChart.
			LineChartAppearance lineApp = this.ChartComponent.GetChartAppearance( ChartAppearanceTypes.Line) as LineChartAppearance;
			LineStyle lnStyle = new LineStyle(
				lineApp.StartStyle,
				lineApp.EndStyle,
				lineApp.DrawStyle
			);
			lnStyle.MidPointAnchors = false;
			Polyline mavgLine = new Polyline( this.movAvg, lnStyle);

			// Set the drawing color and style properties associate with
			// my moving average line.
			mavgLine.Chart =  ChartType.LineChart;
			mavgLine.PE.StrokeWidth = lineApp.Thickness;
			mavgLine.drawColor = this.mavgColor;
			mavgLine.fillColor = this.mavgColor;
			mavgLine.HighLightLine = lineApp.HighLightLines;
				
			// Add to the Scene Graph.
			scene.Add( mavgLine);	
		}

		/// <summary>
		/// Calculates the Simple Moving Average for Layer.  Both the moving
		/// average data (dataspace) and the chart-mapped coordinates (chartspace)
		/// are calculated here.
		/// </summary>
		/// <param name="dp">
		/// A series of DataPoint representing the moving average that will
		/// be plotted upon the scene graph as an overlay layer.
		/// </param>
		protected void calcSMA( DataPoint[] dp)
		{
			// Initialize advance axies for dataspace-to-chartspace
			// translation.
			IAdvanceAxis X = null, Y = null;
			InitAdvanceAxis( ref X, ref Y);

			// Keep frequently referred-to counts of rows & columns.
			int rowCount = this.GetData().GetRowCount();
			int colCount = this.GetData().GetColumnCount();

			// Ideally, this could be factored out and only calculated when IChartData changes.
			double[] smaData = new double[ rowCount];

			Debug.Assert( 
				( daysMovAvg < rowCount ),
				"daysMovAvg < this.GetData().GetRowCount()",
				"This scenario should have been prevented by the Days property validation."
			);

			// Preload the lead-in historical data; this early in
			// the calculation the moving average is unstable and
			// must not be plotted.
			double smaValue = 0.0;
			for ( int n = 0; n < this.daysMovAvg; ++n)
			{
				smaData[ n] = this.GetData().GetValue( n, 4) / (double)( this.Days);
				smaValue += smaData[ n];
			}

			// By virtue of preloading the lead-in data, I've already
			// calculated the first Moving Average data point.
			dp[ 0] = new DataPoint( (int)X.Map( (int)( this.Days) - 1), (int)Y.Map( smaValue));
			dp[ 0].Row = (int)( this.Days);
			dp[ 0].Column = 4;
			//dp[ 0].HitMe = true;
			dp[ 0].Value = this.GetData().GetValue( (int)( this.Days), 4);
	
			// Calculate the Simple Moving Average by rotating in the
			// successor and rotating out the (daysMovAvg)th predecessor.
			for ( int m = (int)( this.Days), index = 1; m < rowCount; ++m, ++index)
			{
				double val = this.GetData().GetValue( m, 4);

				// Rotate-in, rotate-out moving average calculation.
				smaData[ m] = val / (double)( this.Days);
				smaValue += smaData[ m];
				smaValue -= smaData[ m - (int)( this.Days)];

				// Create a point mapped into the chartspace.
				dp[ index] = new DataPoint( (int)X.Map( m), (int)Y.Map( smaValue));
				dp[ index].Row = m;
				dp[ index].Column = 4;
				//dp[ index].HitMe = true;
				dp[ index].Value = val;
			}
		}
	}
}
